// x86_branch_LRAndR.C
// See the .h file for comments

#include "x86_branch_LRAndR.h"
#include "x86_instr.h"
#include "x86_reg_set.h"
#include "directToMemoryEmitter.h"
#include "funkshun.h"
#include "usualJump.h"

branch_LRAndR::branch_LRAndR(kptr_t iLaunchSiteAddr,
                             const instr_t *ioriginsn,
                             const regset_t *iFreeIntRegsBeforeInsn,
                             const regset_t *iFreeIntRegsAfterInsn,
                             SpringBoardHeap &isbheap) : 
   LRAndR(iLaunchSiteAddr, 
          ioriginsn, 
          iFreeIntRegsBeforeInsn, 
          iFreeIntRegsAfterInsn,
          isbheap)
{
}

unsigned branch_LRAndR::conservative_numInsnBytesForRAndR() const {
   unsigned numbytes = originsn->getNumBytes();

   if(originsn->isCall()) {
      /* relocated calls may schedule, so we need to guard the patch
         with an entry counter that is incremented in the instrumentation
         and decremented before the return jump after the relocated call.
         in addition, we store the counter at the end of the patch, so we
         need to account for that space as well. */
      numbytes += 24; // counter space (4) + inc code (8) + dec code (12)
   }
   else if(numbytes < 5)
      // if orig insn was a short jump, it will be turned into a long jump,
      // so we need to account for that (max size of long jump is 7 bytes)
      numbytes = 7;
  
   return numbytes + 5; // relocated insn + jmp to [lauchSite + insn.size()]
}

void branch_LRAndR::
registerPatchAddrAndCreateLauncher(kptr_t ipatchAddr, unsigned ipatchLen,
                                   const instr_t *whenAllDoneRestoreToThisInsn) {
   setPatch(ipatchAddr, ipatchLen);
   setLauncherToAnyAnnulledLauncher(patchAddr, whenAllDoneRestoreToThisInsn);
}

void branch_LRAndR::emitStuffAfterPreInsnSnippets(directToMemoryEmitter_t &em,
                                                  const function_t &fn,
                                                  const instr_t::controlFlowInfo*
                                                ) const {
   const unsigned initialNumInsnBytesEmitted = em.numInsnBytesEmitted();
   kptr_t currInsnAddr = patchAddr + em.numInsnBytesEmitted();

   // Basically, we just need to emit the relocated original insn and a 
   // jmp to the insn following the original insn

   x86_instr *i = (x86_instr*)originsn->relocate(launchSiteAddr, currInsnAddr);
   assert(i != NULL);
   em.emit(i);

   if(originsn->isCall()) {
      i = new x86_instr(x86_instr::push, x86_reg::eAX);
      em.emit(i);
      i = new x86_instr(x86_instr::pushfd);
      em.emit(i);
      i = new x86_instr(x86_instr::mov, patchAddr + patchLen - 4, 
                        x86_reg::eAX);
      em.emit(i);
      i = new x86_instr(x86_instr::atomic_dec, x86_reg::eAX);
      em.emit(i);
      i = new x86_instr(x86_instr::popfd);
      em.emit(i);
      i = new x86_instr(x86_instr::pop, x86_reg::eAX);
      em.emit(i);
   }

   currInsnAddr = patchAddr + em.numInsnBytesEmitted();
   int32_t next_insn_disp = launchSiteAddr + originsn->getNumBytes() - 
                            (currInsnAddr + 5);
   i = new x86_instr(x86_instr::jmp, next_insn_disp); 
   em.emit(i);

   assert(em.numInsnBytesEmitted() - initialNumInsnBytesEmitted <= conservative_numInsnBytesForRAndR());
}

