// freeUpCodePatchInternalEvent.h

#ifndef _FREE_UP_CODE_PATCH_INTERNAL_EVENT_H_
#define _FREE_UP_CODE_PATCH_INTERNAL_EVENT_H_

#include "util/h/mrtime.h"
#include "internalEvent.h"
#include "patchHeap.h"
#include "kernelDriver.h"
#include "loop.h" // doOneTimeInternalEventInTheFuture()

extern kernelDriver *global_kernelDriver;

class freeUpCodePatchInternalEvent : public internalEvent {
 private:
   patchHeap &thePatchHeap;

   kptr_t patchAddr;
   unsigned patchNumBytes;
   bool counterProtected;
   
 public:
   freeUpCodePatchInternalEvent(mrtime_t when,
                                patchHeap &ipatchHeap,
                                kptr_t ipatchAddr,
                                unsigned ipatchNumBytes,
                                bool iverboseTiming,
                                bool icounterProtected) :
      internalEvent(when, iverboseTiming),
      thePatchHeap(ipatchHeap),
      patchAddr(ipatchAddr),
      patchNumBytes(ipatchNumBytes),
      counterProtected(icounterProtected) {
   }
   virtual ~freeUpCodePatchInternalEvent() {}

   void doNow(bool verboseTiming) const {
      const mrtime_t startTime = verboseTiming ? getmrtime() : 0;

#ifdef i386_unknown_linux2_4 
      if(counterProtected) {
         kptr_t ctr_addr = patchAddr + patchNumBytes - 4;
         unsigned the_patch_count = global_kernelDriver->peek1Word(ctr_addr);
         if(the_patch_count == 0) {
            thePatchHeap.free(patchAddr, patchNumBytes);
         }
         else {
            // there's some thread still executing inside the patch, delay
            // patch free for 10 more secs. we have to create a new event
            // since the code in loop.C always assumes doNow completes 
            // successfully and thus deletes the event
            internalEvent *e = new freeUpCodePatchInternalEvent(getmrtime() + 10000000000ULL,
                                                                thePatchHeap,
                                                                patchAddr,
                                                                patchNumBytes,
                                                                verboseTiming,
                                                                true);
            assert(e);
            doOneTimeInternalEventInTheFuture(e); // loop.C
         }
      }
      else
#endif
         thePatchHeap.free(patchAddr, patchNumBytes);

      if (verboseTiming) {
         const mrtime_t totalTime = getmrtime() - startTime;
         
         cout << "delayed free-up-code-patch took "
              << totalTime/1000 << " usecs." << endl;
      }
   }
};

#endif
