// ari_2.5.1.C

#include "ari.h"
#include <sys/cmn_err.h> // cmn_err()
#include <sys/modctl.h>
#include <string.h>
#include <sys/kmem.h> // kmem_alloc()
//#include <sys/kobj.h> // struct module
extern "C" int copyout(caddr_t, caddr_t, size_t);
extern "C" int copyin(const void*, void*, size_t);

#include "vm/seg_kmem.h" // from uts/sun
#include <sys/debug.h> // ASSERT
#include "vm/hat.h"
#include "sys/memlist.h"

#include <sys/ddi.h>
#include <sys/sunddi.h>

void ari_hat_chgprot(struct as *as, caddr_t addr, unsigned len, unsigned vprot) {
   ASSERT(AS_LOCK_HELD(as, &as->a_lock));

   struct hat *hat;
   for (hat = as->a_hat; hat != NULL; hat = hat->hat_next) {
      cmn_err(CE_CONT, "ari: here is a hat\n");
      HATOP_CHGPROT(hat, as, addr, len, vprot); // sfmmu_chgprot?
   }
}

int ari_segkmem_setprot(struct seg *seg, caddr_t addr, unsigned len, unsigned prot) {
   extern struct seg_ops segkmem_ops;
   

   if (seg->s_as != &kas) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 1\n");
      return -1;
   }
   else if (len > seg->s_size) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 2\n");
      return -1;
   }
   else if (addr > addr + len) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 3\n");
      return -1;
   }
   else if (addr < seg->s_base || addr + len > seg->s_base + seg->s_size) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 4\n");
      return -1;
   }
   else if (seg->s_ops != &segkmem_ops) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 5\n");
      return -1;
   }
   else if (!RW_READ_HELD(&seg->s_as->a_lock)) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 6\n");
      return -1;
   }

   
   if (addr < seg->s_base || (addr + len) > (seg->s_base + seg->s_size)) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 7\n");
      return -1;
   }

   if (prot == 0) {
      cmn_err(CE_CONT, "ari_segkmem_setprot assert failure 8\n");
      return -1;
   }

   ari_hat_chgprot(seg->s_as, addr, len, prot);
      
   return 0;
}

int ari_as_setprot(struct as *as, caddr_t addr, u_int size, u_int prot) {
   cmn_err(CE_CONT, "ari welcome to ari_as_setprot\n");

   caddr_t pageaddr = (caddr_t)((unsigned)addr & PAGEMASK);
   unsigned roundupsize = (((unsigned)(addr + size) + PAGEOFFSET) & PAGEMASK) -
         (unsigned)pageaddr;

   cmn_err(CE_CONT, "input params: addr=%x size=%u\n", (unsigned)addr, size);
   cmn_err(CE_CONT, "rounded params: addr=%x size=%u\n", (unsigned)pageaddr,
           roundupsize);

   if (pageaddr + roundupsize < pageaddr)
      return -1;
   
   AS_LOCK_ENTER(as, &as->a_lock, RW_READER);

   struct seg *seg = as_segat(as, pageaddr);
   if (seg == &ktextseg)
      cmn_err(CE_CONT, "seg is ktextseg, as expected\n");
   else
      cmn_err(CE_CONT, "seg isn't ktextseg!\n");

   if (0 == ari_segkmem_setprot(seg, pageaddr, roundupsize,
//                            PROT_READ | PROT_EXEC | PROT_WRITE))
                            PROT_READ | PROT_EXEC))
      cmn_err(CE_CONT, "ari segkmem_setprot seems to have succeeded!\n");
   else
      cmn_err(CE_CONT, "ari segkmem_setprot seems to have failed\n");
   
   AS_LOCK_EXIT(as, &as->a_lock);
   
   cmn_err(CE_CONT, "ari leaving ari_as_setprot\n");

   return 0;
}


int kerninst_segkmem(void *vaddr, int len) {
   cmn_err(CE_CONT, "welcome to kerninst_segkmem for addr %x, len=%d\n",
           (unsigned)vaddr, len);

   // /usr/include/vm/seg.h has the public portion of the seg structure
   const caddr_t ktextseg_base = ktextseg.s_base;
   const u_int   ktextseg_size = ktextseg.s_size;
   
   cmn_err(CE_CONT, "ktextseg base addr is %x\n", (unsigned)ktextseg_base);
   cmn_err(CE_CONT, "ktextseg size is %u\n", ktextseg_size);

   const u_int pfnum = hat_getkpfnum((caddr_t)vaddr);
   cmn_err(CE_CONT, "pfnum of %x is %x\n", (unsigned)vaddr, pfnum);

   cmn_err(CE_CONT, "note that PROT_READ is %x\n", PROT_READ);
   cmn_err(CE_CONT, "PROT_EXEC is %x; PROT_WRITE is %x\n", PROT_EXEC, PROT_WRITE);

   cmn_err(CE_CONT, "calling segkmem_getprot on %x\n", (unsigned)vaddr);
   u_int protv[100];

// The following 2 lines commented out 01 oct 98 since doesn't compile for sol 2.6
//   if (0 != ari_segkmem_getprot(&ktextseg, (caddr_t)vaddr, 0, protv))
//      cmn_err(CE_WARN, "ari_segkmem_getprot!!!\n");
   
   // do your own gdb printing here!
   cmn_err(CE_CONT, "protection seems to be: 0x%x\n",
           protv[0]);

   // now try to change the protection!
   // we'd like to call the standard segkmem_setprot(), but it's declared as static.
   // So we use as_setprot(), which is probably better anyway (though slower) because
   // it properly locks the as.
//   cmn_err(CE_CONT, "trying to change the protection to 0x7 now (deferred)!\n");

//   if (0 != ari_as_setprot(&kas, (caddr_t)vaddr, len,
//                       PROT_READ | PROT_WRITE | PROT_EXEC))
//      cmn_err(CE_CONT, "change of protection FAILED!\n");
//   else
//      cmn_err(CE_CONT, "change of protection SUCCEEDED!\n");

   unsigned physpagenum = hat_getkpfnum((caddr_t)vaddr);
   cmn_err(CE_CONT, "vaddr=%x; phys page num=%x\n",
           (unsigned)vaddr, (unsigned)physpagenum);

//    cmn_err(CE_CONT, "calling segkmem_mapin\n");
//    delay(drv_usectohz(1000000)); // 1 sec
//    void *mypage = kmem_alloc(MMU_PAGESIZE, KM_SLEEP);
//    
//    segkmem_mapin(&kvseg, (caddr_t)mypage, MMU_PAGESIZE,
//                  PROT_READ | PROT_WRITE, hat_getkpfnum((caddr_t)vaddr),
//                  HAT_NOCONSIST);
//
//    unsigned page_offset = (unsigned)vaddr & MMU_PAGESIZE;
//    cmn_err(CE_CONT, "vaddr=%x; page_offset=%x; MMU_PAGESIZE=%u\n",
//            (unsigned)vaddr, (unsigned)page_offset, (unsigned)MMU_PAGESIZE);
//
//    cmn_err(CE_CONT, "offset from page start is %u\n", page_offset);
//    
//    
//
//    cmn_err(CE_CONT, "calling segkmem_mapout\n");
//    delay(drv_usectohz(1000000)); // 1 sec
//    segkmem_mapout(&kvseg, (caddr_t)mypage, MMU_PAGESIZE);
//
//    kmem_free(mypage, MMU_PAGESIZE);
   
   cmn_err(CE_CONT, "leaving kerninst_segkmem\n");

   return 0;
}

